#ifndef _COMPOSITIONANDREJECTION_H_
#define _COMPOSITIONANDREJECTION_H_

#include "StochasticSimulator.h"

#include <vector>
#include <cmath>

#include "XPlatformFix.h"
#include TR1(random)
#include TR1(unordered_map)

class CRPosition {
public:
	CRPosition() : group(-1), slot(-1) {}
	CRPosition(int g, int s) : group(g), slot(s) {}
	int getGroup() const {return group;}
	int getSlot() const {return slot;}
private:
	int group;
	int slot;
};

template<typename EVENT>
class CompositionAndRejection : public StochasticSimulator<EVENT> {
	typedef std::vector<EVENT> Events;
	class Group {
	public:
		Group(double pmin, double pmax) : max(pmax), min(pmin), pg(0.0) {}
		Events events;
		double pg;

		// Do not modify these after initialization
		double max;
		double min;
	};
	typedef typename EVENT::HashType HashType;
	typedef std::vector<Group> Groups;
	typedef std::tr1::unordered_map<HashType, CRPosition> PositionMap;
	typedef std::tr1::unordered_map<int, int> GroupMap;

public:
	// TODO : (Michael) Remove need to specify max... since we create groups on
	// the fly max is uneeded as we don't create all necessary groups at start
	CompositionAndRejection(double min, double max) : pmax(max), pmin(min), ps(0.0) {
		engine.seed(time(0));
	}

	virtual ~CompositionAndRejection() {
	}

	virtual bool ContainsEvent(const EVENT& event) const {
		PositionMap::const_iterator pos = collection.find(event.getHash());
		return pos != collection.end();
	}

	virtual bool ModifyEvent(EVENT event) {
		PositionMap::const_iterator pos = collection.find(event.getHash());
		if(pos != collection.end()) {
			// Erase and add take care of adjusting probabilities
			EraseEvent(event);
			AddEvent(event);
			return true;
		} else {
			// TODO: (Michael) log modification on nonexistant event
			return false;
		}
	}

	virtual bool AddEvent(const EVENT& event) {
		HashType hash = event.getHash();
		PositionMap::const_iterator pos = collection.find(hash);
		if(pos == collection.end()) {
			// Retrieve our group
			unsigned int selection = getGroup(event.getProbability());
			GroupMap::const_iterator position = groupPositions.find(selection);
			if(position == groupPositions.end()) {
				// if group does not exist then we create it
				createGroup(selection);
				position = groupPositions.find(selection);
			}
			Group& group = events[position->second];

			// Add event to data structues
			collection.insert(std::make_pair(hash, CRPosition(position->second, group.events.size())));
			group.events.push_back(event);

			// Adjust probabilities
			double adjustment = event.getProbability();
			group.pg += adjustment;
			ps += adjustment;

			return true;
		} else {
			// TODO: (Michael) log trying to add more than one event at a coord
			return false;
		}
	}

	virtual bool EraseEvent(const EVENT& event) {
		HashType hash = event.getHash();
		PositionMap::iterator pos = collection.find(hash);
		if(pos != collection.end()) {
			CRPosition position = pos->second;
			Group& group = events[position.getGroup()];

			// The positions of swapee and erasee within the vector
			int erasee = position.getSlot();
			int swapee = group.events.size() - 1;

			// If erasee is not at the end of the vector
			if(erasee != swapee) {
				// Update the location of swapee
				HashType swapeeHash = group.events[swapee].getHash();
				PositionMap::iterator swapeePos = collection.find(swapeeHash);
				swapeePos->second = pos->second;

				// Change swapee to his new location
				std::swap(group.events[erasee], group.events[swapee]);
			}

			// Adjust probabilities
			// Quick note swapee no longer points to swapee but to erasee
			double adjustment = group.events[swapee].getProbability();
			group.pg -= adjustment;
			ps -= adjustment;

			// These checks are here just in case we have rounding errors
			if(ps < 0.0) {
				ps = 0.0;
			}
			if(group.pg < 0.0) {
				group.pg = 0.0;
			}

			// Erase event
			collection.erase(hash);
			group.events.pop_back();

			return true;
		} else {
			// We do not have the event
			// TODO: (Michael) log attempting to remove nonexistant event
			return false;
		}
	}

	virtual const EVENT& SampleEvent(void) {
		// Pick a group
		double groupChoice = randomDouble() * ps;

		// Find our group.. can be done in O(log(g)) with a binary search
		// but then we lose the ability to create only the groups we use
		double runningTotal = 0.0;
		typename Groups::const_iterator group = events.end();
		if(ps >= 0.0) {
			for(typename Groups::const_iterator it = events.begin(); it != events.end(); ++it) {
				runningTotal += it->pg;
				if(groupChoice < runningTotal) {
					group = it;
					break;
				}
			}
		}

		if(group != events.end()) {
			bool found = false;
			int x = 0;
			double y = 0.0;
			while(!found) {
				// x from [0, events.size) in int
				x = static_cast<int>(randomDouble() * group->events.size());
				y = randomDouble() * group->max;
				if(y <= group->events[x].getProbability()) {
					found = true;
				}
			}
			return group->events[x];
		} else {
			// TODO : (Michael) log that we are not choosing an event
			return emptyEvent;
		}
	}

	virtual double TimeLapsed(double fhat) {
		double prdt = randomDouble();
		double rtotal = fhat * ps; 
		return std::log(rtotal / prdt) / rtotal;
	}

private:
	void createGroup(int number) {
		double current(pmin);
		double next(current * 2);
		for(int i(1); i <= number; ++i) {
			current = next;
			next *= 2;
		}
		groupPositions.insert(std::make_pair(number, events.size()));
		events.push_back(Group(current, next));
	}

	int getGroup(double p) const {
		int group(0);
		std::frexp(p / pmin, &group);
		return group - 1;
	}

	int maxGroups() const {
		return getGroup(pmax);
	}

	double randomDouble() {
		return distribution(engine);
	}

	GroupMap groupPositions;
	PositionMap collection;
	Groups events;

	double pmax;
	double pmin;
	double ps;

	std::tr1::mt19937 engine;
	std::tr1::uniform_real_distribution<double> distribution;

	EVENT emptyEvent;
};

#endif //_COMPOSITIONANDREJECTION_H_
