/////////////////////////////////////////////////////////////////////////////
// Name:        TASTilePanel.cpp
// Purpose:     Defines the canvas on which the 2-D tile space is drawn
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TASDockers.h"
#include "TASTilePanel.h"
#include "TASApp.h"
#include <wx/clipbrd.h>
#include <wx/regex.h>
#include <wx/dcbuffer.h>
#include "TASDebug.h"

//  Tile type instance to be used as placeholder for blanks
TASTileType gTileTypeNull(2);


bool TASIntVect::Contains(int i)
{
  for (size_t j=0; j<size(); j++) {
    if (at(j) == i) return true;
  }
  return false;
}


//////////////////////////////////////////////////////////////////////////
//  Implementation of TASTileEditor which is the main panel for tiletype
//  editing

IMPLEMENT_CLASS(TASTileEditor, wxScrolledWindow)
BEGIN_EVENT_TABLE(TASTileEditor, wxScrolledWindow)
  EVT_PAINT(TASTileEditor::OnPaint)
  EVT_SIZE(TASTileEditor::OnSize)
  EVT_MOUSE_EVENTS(TASTileEditor::OnMouse)
  EVT_MENU(ID_TAS_TILEED_EDIT,          TASTileEditor::OnEdit)
  EVT_MENU(ID_TAS_TILEED_INSERT,        TASTileEditor::OnInsert)
  EVT_MENU(ID_TAS_TILEED_CUT,           TASTileEditor::OnCut)
  EVT_MENU(ID_TAS_TILEED_COPY,          TASTileEditor::OnCopy)
  EVT_MENU(ID_TAS_TILEED_PASTE,         TASTileEditor::OnPaste)
  EVT_MENU(ID_TAS_TILEED_DELETE,        TASTileEditor::OnDelete)
  EVT_MENU(ID_TAS_TILEED_ROTATE90,      TASTileEditor::OnRotate)
  EVT_MENU(ID_TAS_TILEED_ROTATE180,     TASTileEditor::OnRotate)
  EVT_MENU(ID_TAS_TILEED_ROTATE270,     TASTileEditor::OnRotate)
  EVT_MENU(ID_TAS_TILEED_FINDBINDING_N, TASTileEditor::OnFindBinding)
  EVT_MENU(ID_TAS_TILEED_FINDBINDING_E, TASTileEditor::OnFindBinding)
  EVT_MENU(ID_TAS_TILEED_FINDBINDING_S, TASTileEditor::OnFindBinding)
  EVT_MENU(ID_TAS_TILEED_FINDBINDING_W, TASTileEditor::OnFindBinding)
  EVT_MENU(ID_TAS_TILEED_FIND,          TASTileEditor::OnFind)
  EVT_MENU(ID_TAS_TILEED_SELECTUSED,    TASTileEditor::OnSelectUsed)
  EVT_MENU(ID_TAS_TILEED_SELECTUNUSED,  TASTileEditor::OnSelectUsed)
  EVT_KEY_DOWN(                         TASTileEditor::OnKey)
END_EVENT_TABLE()

TASTileEditor::TASTileEditor(wxWindow *parent, TASEditorFrame *editorFrame, wxWindowID id,
                                 const wxPoint& pos, const wxSize& size, long style, const wxString& name)
                             : wxScrolledWindow(parent, id, pos, size, style, name)
{
  m_pEditorFrame      = editorFrame;
  m_tileSpacing       = 20;
  m_tilesPerRow       = 1;
  m_currTileTypeNum   = -1;
  m_savedTileTypeNum  = -1;
  m_ptPreDragStart    = wxPoint(-1, -1);
  m_bInDrag           = false;
  m_bInSelectDrag     = false;
  m_size3DTiles       = wxSize(180, 240);

  wxSizeEvent evt;
  OnSize(evt);

  SetBackgroundStyle(wxBG_STYLE_CUSTOM);

  SetBackgroundColour(g_pContext->m_colSpaceBkgrnd);
  ClearBackground();

////////////////////////////////////////////////////////
// Unit testing for TASTileTypeSet class
/*
  TASTileType tt1(2);
  tt1.m_sName = "first";
  tt1.m_sLabel = "first label";

  TASTileType tt2(2);
  tt2.m_sName = "second";
  tt2.m_sLabel = "second label";

  TASTileType tt3(2);
  tt3.m_sName = "third";
  tt3.m_sLabel = "third label";

  TASTileTypeSet foo;
  foo.push_back(tt1);
  foo.push_back(tt2);
  foo.push_back(tt3);
  foo.push_back(tt2);

  size_t s = foo.size();

  TASTileType out0 = foo["first"];
  TASTileType out1 = foo["second"];
  TASTileType out2 = foo["fake"];
  TASTileType out3(foo["third"]);
  bool b0 = out3.IsValid();

  TASTileType out4 = foo[0];
  TASTileType out5 = foo[1];
  TASTileType out6 = foo[3];

  bool b1 = out4.IsValid();
  bool b2 = out6.IsValid();

  TASTileType ins1;
  ins1.m_sName = "inserted1";
  ins1.m_sLabel = "first inserted";

  foo.insert(ins1, 1);

  TASTileType out7 = foo[0];
  TASTileType out8 = foo[1];
  TASTileType out9 = foo[2];
  TASTileType out10 = foo[3];

  size_t s2 = foo.size();

  TASTileType stt1 = foo["first"];
  TASTileType stt2 = foo["second"];
  TASTileType stt3 = foo["inserted1"];
  TASTileType stt4 = foo["third"];
  TASTileType stt5 = foo["horseshit"];

  foo.erase(2);

  TASTileType out11 = foo[0];
  TASTileType out12 = foo[1];
  TASTileType out13 = foo[2];
  TASTileType out14 = foo[3];

  size_t s3 = foo.size();

  TASTileType stt6 = foo["first"];
  TASTileType stt7 = foo["second"];
  TASTileType stt8 = foo["inserted"];
  TASTileType stt9 = foo["third"];
  TASTileType stt10 = foo["horseshit"];

  TASTileTypeSet foo2;
  foo2 = foo;
  for (size_t t=0; t<foo2.size(); t++) {
    TASTileType foo2tt = foo2[t];
  }
  TASTileType foo2tt2 = foo2["inserted1"];
  TASTileType foo2tt3 = foo2["first"];
  TASTileType foo2tt4 = foo2["junkjunk"];
  TASTileType foo2tt5 = foo2["third"];
  TASTileType foo2tt6 = foo2["second"];

  foo.clear();
  size_t s4 = foo.size();

  size_t i1 = foo2.IndexOf("first");
  bool bi1 = i1 == TAS_TILE_NOT_FOUND;
  size_t i2 = foo2.IndexOf("sdflsjf");
  bool bi2 = i2 == TAS_TILE_NOT_FOUND;

  const TASTileType* pTT = &foo2["first"];
  wxString sTTLabel = pTT->m_sLabel;
*/
///////////////////////////////////////////////////////////
}

void TASTileEditor::GetTileLocation(int tileNum, int* x, int* y)
{
  if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
    *x = (tileNum % m_tilesPerRow) * (g_tileSize + m_tileSpacing) + 20;
    *y = (tileNum / m_tilesPerRow) * (g_tileSize + m_tileSpacing) + 20;
  } else {
    *x = (tileNum % m_tilesPerRow) * (m_size3DTiles.GetWidth() + m_tileSpacing) + 20;
    *y = (tileNum / m_tilesPerRow) * (m_size3DTiles.GetHeight() + m_tileSpacing) + 20;
  }
}

void TASTileEditor::GetTileRect(int tileNum, wxRect* r)
{
  int tx, ty;
  GetTileLocation(tileNum, &tx, &ty);
  int z = 1;
  wxDouble x, y, w, h;
  x = z-(m_tileSpacing/2);
  y = z-(m_tileSpacing/2);
  if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
    w = (g_tileSize + m_tileSpacing - (2*z));
    h = (g_tileSize + m_tileSpacing - (2*z));
    r->x = tx + x;
    r->y = ty + y;
    r->width = w;
    r->height = h;
  } else {
//TODO: MJP - need to implement this!!
  }
}

int TASTileEditor::GetTileFromLocation(int x, int y)
{
  int retVal = -1;

  int xSize = g_tileSize;
  int ySize = g_tileSize;

  if (g_pContext->m_iTASDimMode == TAS_3D_MODE) {
    xSize = m_size3DTiles.GetWidth();
    ySize = m_size3DTiles.GetHeight();
  }

  int column = (x - 20) / (xSize + m_tileSpacing);
  int row    = (y - 20) / (ySize + m_tileSpacing);

  if (column < m_tilesPerRow) {
    retVal = (row * m_tilesPerRow) + column;
  }

  return (retVal >= (int)g_pEditorTileset->size() ? -1 : retVal);
}

int TASTileEditor::GetTileFromGridLoc(int row, int col)
{
  return (row * m_tilesPerRow) + col;
}

void TASTileEditor::OnPaint(wxPaintEvent& WXUNUSED(event))
{
  wxAutoBufferedPaintDC dcAutoBuff(this);
  DoPrepareDC(dcAutoBuff);

  SetBackgroundColour(g_pContext->m_colSpaceBkgrnd);
//  dcAutoBuff.SetBackground(m_brushOffspace);
  dcAutoBuff.Clear();

  wxBrush brush(g_pContext->m_colSpaceBkgrnd);
  wxPen pen(g_pContext->m_colSpaceBkgrnd);

  // Find Out where the window is scrolled to
  int vbX,vbY;                              // Top left corner of client
  GetViewStart(&vbX,&vbY);
  int xUnit, yUnit;
  GetScrollPixelsPerUnit(&xUnit, &yUnit);
  vbX = vbX * xUnit;
  vbY = vbY * yUnit;

  TASCoords coordsCurr;
  int vX,vY,vW,vH;                          // Dimensions of client area in pixels
  wxRegionIterator upd(GetUpdateRegion());  // get the update rect list

  int xSize = g_tileSize;
  int ySize = g_tileSize;

  if (g_pContext->m_iTASDimMode == TAS_3D_MODE) {
    xSize = m_size3DTiles.GetWidth();
    ySize = m_size3DTiles.GetHeight();
  }

  while (upd)
  {
    vX = vbX + upd.GetX();
    vY = vbY + upd.GetY();
    vW = upd.GetW();
    vH = upd.GetH();

    // Repaint this rectangle
    dcAutoBuff.SetBrush(brush);
    dcAutoBuff.SetPen(pen);
    dcAutoBuff.DrawRectangle(vX, vY, vW, vH);

    // Find start/end row/column
    int startColumn = (vX - 20) / (xSize + m_tileSpacing);
    int startRow    = (vY - 20) / (ySize + m_tileSpacing);
    int startTile   = GetTileFromGridLoc(startRow, startColumn);

    int endColumn = (vX + vW - 20) / (xSize + m_tileSpacing);
    int endRow    = (vY + vH - 20) / (ySize + m_tileSpacing);
    int endTile   = GetTileFromGridLoc(endRow, endColumn);

    for (int i = startTile; i<= endTile; i++) {
      DrawTileEx(&dcAutoBuff, i);
    }
    upd++;
  }
}

void TASTileEditor::OnSize(wxSizeEvent& WXUNUSED(event))
{
  int w, h;
  GetClientSize(&w, &h);
  if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
    m_tilesPerRow = w / (g_tileSize + m_tileSpacing);
    SetVirtualSize((m_tilesPerRow * (g_tileSize + m_tileSpacing)), (((g_pEditorTileset->size() / m_tilesPerRow) + 1) * (g_tileSize + m_tileSpacing)));
  } else {
    m_tilesPerRow = w / (m_size3DTiles.GetWidth() + m_tileSpacing);
    SetVirtualSize((m_tilesPerRow * (m_size3DTiles.GetWidth() + m_tileSpacing)), (((g_pEditorTileset->size() / m_tilesPerRow) + 1) * (m_size3DTiles.GetHeight() + m_tileSpacing)));
  }
  SetScrollRate(20, 20);

  Refresh();
}

void TASTileEditor::SelectTile(int tileNum)
{
  if (tileNum >= 0) {
    m_selectedList.push_back(tileNum);

    wxRect r;
    GetTileRect(tileNum, &r);
    Refresh(true, &r);

    if (m_pEditorFrame && m_pEditorFrame->m_pTTWin) {
      m_pEditorFrame->m_pTTWin->EnableEditing(true);
    }
  }
}

void TASTileEditor::UnSelectTile(int tileNum)
{
  if (tileNum >= 0) {
    for (TASIntVectIter iter = m_selectedList.begin(); iter != m_selectedList.end(); iter++) {
      if (*iter == tileNum) {
        m_selectedList.erase(iter);
        break;
      }
    }

    wxRect r;
    GetTileRect(tileNum, &r);
    Refresh(true, &r);

    if (m_pEditorFrame && m_pEditorFrame->m_pTTWin) {
      m_pEditorFrame->m_pTTWin->EnableEditing(false);
    }
  }
}

bool TASTileEditor::IsTileSelected(int tileNum)
{
  for (size_t i=0; i<m_selectedList.size(); i++) {
    if (m_selectedList[i] == tileNum) {
      return true;
    }
  }
  return false;
}

bool TASTileEditor::IsOverSelectedTile(wxPoint& pt)
{
  for (size_t i=0; i<m_selectedList.size(); i++) {
    int x, y;
    GetTileLocation(m_selectedList[i], &x, &y);
    if ((pt.x >= x) && (pt.x <= (x + g_tileSize + m_tileSpacing)) &&
        (pt.y >= y) && (pt.y <= (y + g_tileSize + m_tileSpacing)))
    {
      return true;
    }
  }
  return false;
}

void TASTileEditor::OnMouse(wxMouseEvent& event)
{
  wxClientDC dc(this);
  PrepareDC(dc);
  m_ptMousePos = event.GetLogicalPosition(dc);

  int tileNum = GetTileFromLocation(m_ptMousePos.x, m_ptMousePos.y);

  if (event.GetEventType() == wxEVT_LEAVE_WINDOW) {                 //  Cursor left the window
    tileNum = -1;
    if (m_bInDrag) {
      m_bInDrag = false;
      Refresh();
    }
    if (m_bInSelectDrag) {
      m_bInSelectDrag = false;
//MJP: unselect all
    }
  } else if (event.LeftDClick()) {                                  //  Double-click
//MJP: position needs to be modified!!!
//    TASTileTypeWin ttw(this, m_currTileTypeNum, m_ptMousePos);
//    ttw.ShowModal();
  } else if (event.Dragging() && event.LeftIsDown()) {              //  Doing dragging
    if (event.m_shiftDown) {                                        //    Selection
      if (!m_bInSelectDrag) {
        m_ptDragStart = m_ptMousePos;
      }
      m_bInSelectDrag = true;

      //  Calculate the set of tiles that are currently selected
      int t1 = GetTileFromLocation(m_ptDragStart.x, m_ptDragStart.y);
      int t2 = GetTileFromLocation(m_ptDragStart.x, m_ptMousePos.y);
      int t3 = GetTileFromLocation(m_ptMousePos.x,  m_ptDragStart.y);

      int rowStart    = t1 / m_tilesPerRow;
      int rowFinish   = t2 / m_tilesPerRow;
      int colStart    = t1 % m_tilesPerRow;
      int colFinish   = t3 % m_tilesPerRow;

      TASIntVect oldSel = m_selectedList;
      m_selectedList.clear();
      int r1 = (rowStart < rowFinish) ? rowStart : rowFinish;
      int r2 = (r1 == rowStart) ? rowFinish : rowStart;
      int c1 = (colStart < colFinish) ? colStart : colFinish;
      int c2 = (c1 == colStart) ? colFinish : colStart;
      for (int r=r1; r<=r2; r++) {
        for (int c=c1; c<=c2; c++) {
          m_selectedList.push_back(GetTileFromGridLoc(r, c));
        }
      }

      //  Redraw tiles that have become unselected
      for (size_t o=0; o<oldSel.size(); o++) {
        if (!m_selectedList.Contains(oldSel[o])) {
          wxRect r;
          GetTileRect(oldSel[o], &r);
          Refresh(true, &r);
        }
      }
      //  Redraw selected tiles
      for (size_t n=0; n<m_selectedList.size(); n++) {
          wxRect r;
          GetTileRect(m_selectedList[n], &r);
          Refresh(true, &r);
      }
    } else {                                                        //    Moving tile(s)
      if (!m_bInDrag && IsOverSelectedTile(m_ptMousePos)) {
        if (m_ptPreDragStart == wxPoint(-1, -1)) {
          m_ptPreDragStart = m_ptMousePos;
        } else if ((abs(m_ptPreDragStart.x - m_ptMousePos.x) >= 5) || (abs(m_ptPreDragStart.y - m_ptMousePos.y) >= 5)){
          m_ptDragStart = m_ptPreDragStart;
          m_ptPreDragStart = wxPoint(-1, -1);
          m_bInDrag = true;
        }
      }

      if (m_bInDrag) {

        //  Redraw all tiles that are being passed over
//MJP: this will only handle cleanup drawing for dragging one tile at the moment (and it doesn't even work that well)
        int dist = g_tileSize + m_tileSpacing;
        int tn = GetTileFromLocation(m_ptMousePos.x, m_ptMousePos.y);
        DrawTileEx(&dc, tn);
        tn = GetTileFromLocation(m_ptMousePos.x - dist, m_ptMousePos.y - dist);
        if (tn >= 0) DrawTileEx(&dc, tn);
        tn = GetTileFromLocation(m_ptMousePos.x - dist, m_ptMousePos.y + dist);
        if (tn >= 0) DrawTileEx(&dc, tn);
        tn = GetTileFromLocation(m_ptMousePos.x + dist, m_ptMousePos.y - dist);
        if (tn >= 0) DrawTileEx(&dc, tn);
        tn = GetTileFromLocation(m_ptMousePos.x + dist, m_ptMousePos.y + dist);
        if (tn >= 0) DrawTileEx(&dc, tn);

        //  Draw the tiles being dragged
        for (size_t i=0; i<m_selectedList.size(); i++) {
          DrawTileEx(&dc, m_selectedList[i], (m_ptMousePos.x - m_ptDragStart.x), (m_ptMousePos.y - m_ptDragStart.y));
        }
      }
    }
  } else if (event.LeftUp()) {                                      //  Left button coming up
    if (m_bInDrag) {                                                //    Ending tile drag
      m_bInDrag = false;
      PutTileTypesOnClipboard(m_selectedList, true);
      m_savedTileTypeNum = tileNum;
      wxCommandEvent evt;
      OnPaste(evt);
    } else if (m_bInSelectDrag) {                                   //    Ending tile selection drag
      m_bInSelectDrag = false;
    } else {                                                        //    Ending single left click
      bool bIsSelected = IsTileSelected(tileNum);
      while (m_selectedList.size() > 0) {
        UnSelectTile(m_selectedList[0]);
      }
      if ((tileNum >= 0) && (!bIsSelected)) {
        SelectTile(tileNum);
        m_pEditorFrame->m_pTTWin->SetCurrentTileType(tileNum);
      }
    }
  } else if (event.RightIsDown()) {                                 //  Holding down right button - show menu
    m_savedTileTypeNum = tileNum;
    wxMenu* pMenu = new wxMenu();
    BuildEditMenu(pMenu);
    PopupMenu(pMenu);
    delete pMenu;
  }

  if (tileNum != m_currTileTypeNum) {
    if (m_selectedList.size() <= 0) {
      m_pEditorFrame->m_pTTWin->SetCurrentTileType(tileNum);
    }
    int prevNum = m_currTileTypeNum;
    m_currTileTypeNum = tileNum;
    wxRect r;
    GetTileRect(prevNum, &r);
    Refresh(true, &r);
    if (tileNum >= 0) {
      wxRect r;
      GetTileRect(tileNum, &r);
      Refresh(true, &r);
    }
  }
}

void TASTileEditor::BuildEditMenu(wxMenu* pMenu)
{
  pMenu->Append(ID_TAS_TILEED_EDIT, _T("Edit"));
  pMenu->AppendSeparator();
  pMenu->Append(ID_TAS_TILEED_INSERT, _T("Insert New\tCtrl-I"));
  pMenu->AppendSeparator();
  pMenu->Append(ID_TAS_TILEED_CUT,    _T("Cut\tCtrl-X"));
  pMenu->Append(ID_TAS_TILEED_COPY,   _T("Copy\tCtrl-C"));
  pMenu->Append(ID_TAS_TILEED_PASTE,  _T("Paste\tCtrl-V"));
  pMenu->Append(ID_TAS_TILEED_DELETE, _T("Delete\tDel"));
  pMenu->AppendSeparator();
  if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
    wxMenu* pRotMenu = new wxMenu();
    pRotMenu->Append(ID_TAS_TILEED_ROTATE90,  _T(" 90 degrees"));
    pRotMenu->Append(ID_TAS_TILEED_ROTATE180, _T("180 degrees"));
    pRotMenu->Append(ID_TAS_TILEED_ROTATE270, _T("270 degrees"));
    pMenu->AppendSubMenu(pRotMenu, _T("Rotate (CW)"));
    pMenu->AppendSeparator();
    pMenu->Append(ID_TAS_TILEED_FIND, _T("Find tiles...\tCtrl-F"));
    wxMenu* pBindMenu = new wxMenu();
    pBindMenu->Append(ID_TAS_TILEED_FINDBINDING_N, _T("North"));
    pBindMenu->Append(ID_TAS_TILEED_FINDBINDING_E, _T(" East"));
    pBindMenu->Append(ID_TAS_TILEED_FINDBINDING_S, _T("South"));
    pBindMenu->Append(ID_TAS_TILEED_FINDBINDING_W, _T(" West"));
    pMenu->AppendSubMenu(pBindMenu, _T("Find tiles that can bind to the.."));
  } else {
    wxMenu* pRotMenu = new wxMenu();
    pRotMenu->Append(ID_TAS_TILEED_ROTATE90,  _T(" 90 degrees"));
    pRotMenu->Append(ID_TAS_TILEED_ROTATE180, _T("180 degrees"));
    pRotMenu->Append(ID_TAS_TILEED_ROTATE270, _T("270 degrees"));
    pMenu->AppendSubMenu(pRotMenu, _T("Rotate (CW)"));
    pMenu->AppendSeparator();
    pMenu->Append(ID_TAS_TILEED_FIND, _T("Find tiles...\tCtrl-F"));
    wxMenu* pBindMenu = new wxMenu();
    pBindMenu->Append(ID_TAS_TILEED_FINDBINDING_N, _T("North"));
    pBindMenu->Append(ID_TAS_TILEED_FINDBINDING_E, _T(" East"));
    pBindMenu->Append(ID_TAS_TILEED_FINDBINDING_S, _T("South"));
    pBindMenu->Append(ID_TAS_TILEED_FINDBINDING_W, _T(" West"));
    pMenu->AppendSubMenu(pBindMenu, _T("Find tiles that can bind to the.."));
  }
//if (m_savedTileTypeNum == 0) {
//if (m_tileTypeSet.GetEquivalentTileType(&m_tileTypeSet[m_savedTileTypeNum], false, 0) != NULL) {
//  pMenu->Append(wxID_ANY, _T("Find equivalent tiletypes"));
//}
  pMenu->AppendSeparator();
  pMenu->Append(ID_TAS_TILEED_SELECTUSED, _T("Select used tiletypes"));
  pMenu->Append(ID_TAS_TILEED_SELECTUNUSED, _T("Select unused tiletypes"));
}

void TASTileEditor::OnEdit(wxCommandEvent& WXUNUSED(event))
{
  if (m_savedTileTypeNum >= 0) {
    while (m_selectedList.size() > 0) {
      UnSelectTile(m_selectedList[0]);
    }
    SelectTile(m_savedTileTypeNum);
    if (m_pEditorFrame && m_pEditorFrame->m_pTTWin) {
      m_pEditorFrame->m_pTTWin->SetCurrentTileType(m_savedTileTypeNum);
    }
  }
}

void TASTileEditor::OnInsert(wxCommandEvent& WXUNUSED(event))
{
  TASTileType newTT(2);
  newTT.m_sName = GenerateTileName(_T("New Tile"));

  size_t index = (size_t)m_savedTileTypeNum;
  if (index != TAS_TILE_NOT_FOUND) {
    g_pEditorTileset->insert(newTT, index);
  } else {
    g_pEditorTileset->push_back(newTT);
    index = g_pEditorTileset->size() - 1;
  }
  g_pEditorTileset->SetSimMod(true);
  SelectTile(index);
  ScrollToCenter(index);
  if (m_pEditorFrame && m_pEditorFrame->m_pTTWin) {
    m_pEditorFrame->m_pTTWin->SetCurrentTileType(index, true);
  }

  Refresh();
}

bool TASTileEditor::PutTileTypesOnClipboard(TASIntVect& ttNumList, bool bDelete)
{
  m_savedList = ttNumList;
  wxString sClipData;
  int numErased = 0;
  for (size_t i=0; i<ttNumList.size(); i++) {
    TASTileType* sTT = g_pEditorTileset->at(ttNumList[i] - numErased);
    sClipData += sTT->Serialize();
    if (bDelete) {
      g_pEditorTileset->erase(ttNumList[i] - numErased);
      numErased++;
    }
  }

  m_sClipboardData = sClipData;
  if (wxTheClipboard->Open()) {
    if (!wxTheClipboard->SetData( new wxTextDataObject(sClipData.c_str()))) {

    }
    wxTheClipboard->Close();
  }
  return true;
}

void TASTileEditor::OnCut(wxCommandEvent& WXUNUSED(event))
{
  if (m_selectedList.size() > 0) {
    g_pEditorTileset->SetSimMod(true);
  }
  PutTileTypesOnClipboard(m_selectedList, true);
  m_selectedList.clear();
  Refresh();
}

void TASTileEditor::OnCopy(wxCommandEvent& WXUNUSED(event))
{
  PutTileTypesOnClipboard(m_selectedList, false);
}

void TASTileEditor::OnPaste(wxCommandEvent& WXUNUSED(event))
{
  wxString sClipText;
  if (wxTheClipboard->Open()) {
    if (wxTheClipboard->IsSupported(wxDF_TEXT)) {
      wxTextDataObject data;
      if (!wxTheClipboard->GetData(data)) {
        sClipText = m_sClipboardData;
      }
      sClipText = data.GetText();
    }
    wxTheClipboard->Close();
  }

  if (sClipText.Len() == 0) {
    sClipText = m_sClipboardData;
  }

  if (sClipText.Len() > 0) {
    bool bDidPaste = false;

    m_selectedList.clear();

    int numInserted = 0;
    wxString sText = sClipText;
    size_t strIndex = 0;
    while (strIndex < sText.Len()) {
      //  Create tiletype object from text on clipboard
      TASTileType ttNew(sText, &strIndex);
      if (ttNew.IsValid()) {
        ttNew.m_sName = GenerateTileName(ttNew.m_sName);

        //  Get insertion location
        size_t index = (size_t)(m_currTileTypeNum + numInserted);
        if (index != TAS_TILE_NOT_FOUND) {
          index++;
        }

        //  Insert the new tiletype
        if ((index == TAS_TILE_NOT_FOUND) ||
            (index >= g_pEditorTileset->size()))
        {
          g_pEditorTileset->push_back(ttNew);
          m_currTileTypeNum = g_pEditorTileset->size() - 1;
        } else {
          g_pEditorTileset->insert(ttNew, index);
          numInserted++;
        }
        bDidPaste = true;
      }
    }
    if (bDidPaste) {
      ScrollToCenter(m_currTileTypeNum + numInserted);
      g_pEditorTileset->SetSimMod(true);
    }
    Refresh();
  }
}

void TASTileEditor::OnDelete(wxCommandEvent& WXUNUSED(event))
{
  if (m_selectedList.size() > 0) {
    wxString sMsg = _T("Are you sure you want to delete the selected tile type?");
    if (m_selectedList.size() > 1) {
      sMsg = wxString::Format(_T("Are you sure you want to delete the %d selected tile types?"), m_selectedList.size());
    }
    wxMessageDialog dlg(this, sMsg, _T("Confirm Delete"), wxICON_QUESTION|wxYES_NO);
    if (dlg.ShowModal() != wxID_YES) {
      return;
    }
    g_pEditorTileset->SetSimMod(true);
  }

  int numErased = 0;
  for (size_t i=0; i<m_selectedList.size(); i++) {
    //TASTileType* sTT = g_pEditorTileset->at(m_selectedList[i] - numErased);
    g_pEditorTileset->erase(m_selectedList[i] - numErased);
    numErased++;
  }

  m_selectedList.clear();
  Refresh();
}

void TASTileEditor::OnRotate(wxCommandEvent& event)
{
  bool bDidRotate = false;
  for (size_t i=0; i<m_selectedList.size(); i++) {
    int tileNum = m_selectedList[i];
    switch (event.GetId()) {
      case ID_TAS_TILEED_ROTATE90:
        g_pEditorTileset->at(tileNum)->Rotate(90);
        bDidRotate = true;
        break;
      case ID_TAS_TILEED_ROTATE180:
        g_pEditorTileset->at(tileNum)->Rotate(180);
        bDidRotate = true;
        break;
      case ID_TAS_TILEED_ROTATE270:
        g_pEditorTileset->at(tileNum)->Rotate(270);
        bDidRotate = true;
        break;
      default:
        break;
    }
    wxRect r;
    GetTileRect(tileNum, &r);
    Refresh(true, &r);
  }
  if (bDidRotate) {
    g_pEditorTileset->SetSimMod(true);
    if (m_pEditorFrame && m_pEditorFrame->m_pTTWin) {
      m_pEditorFrame->m_pTTWin->SetCurrentTileType(m_savedTileTypeNum);
    }
  }
}

///////////////////////////////////////////////////////////////////////

class TASFindTilesWindow : public wxDialog
{
public:
  TASFindTilesWindow(wxWindow* pParent);

  void OnOk(wxCommandEvent& event);

  bool                  m_arrSearchProperties[8];
  wxString              m_arrSearchStrings[8];

private:
  wxBoxSizer*           m_pSizerMain;
  wxBoxSizer*           m_pSizerChecks;
  wxBoxSizer*           m_pSizerTextBoxes;

  wxCheckBox*           m_pArrCheckBoxes[8];
  wxTextCtrl*           m_pArrTextBoxes[8];

  DECLARE_CLASS(TASFindTilesWindow)
  DECLARE_EVENT_TABLE()
};

IMPLEMENT_CLASS(TASFindTilesWindow, wxDialog)
BEGIN_EVENT_TABLE(TASFindTilesWindow, wxDialog)
  EVT_BUTTON(wxID_OK, TASFindTilesWindow::OnOk)
END_EVENT_TABLE()

TASFindTilesWindow::TASFindTilesWindow(wxWindow* pParent)
                      : wxDialog(pParent, wxID_ANY, _T(""), wxDefaultPosition)
{
  wxString sTitle = _T("Find Tiles");
  SetTitle(sTitle);

  m_pSizerMain = new wxBoxSizer(wxVERTICAL);

  m_pSizerMain->AddSpacer(15);

  wxString sMsg;
  sMsg += _T("Select one or more tile properties to search and\n");
  sMsg += _T("enter the string(s) to be searched for within them.\n");

  wxStaticText* pMainText = new wxStaticText(this, wxID_ANY, sMsg);
  m_pSizerMain->Add(pMainText, 0, wxBOTTOM | wxALIGN_CENTER_HORIZONTAL, 10);

  wxBoxSizer* pSizerControls = new wxBoxSizer(wxHORIZONTAL);

  int dirNum = (g_pContext->m_iTASDimMode == TAS_2D_MODE) ? 4 : 6;

  int vertSpaceChecks = 11;
  int vertSpaceText   = 5;

  m_pSizerChecks    = new wxBoxSizer(wxVERTICAL);
  m_pSizerTextBoxes = new wxBoxSizer(wxVERTICAL);

  m_pSizerChecks->AddSpacer(5);

  m_pArrCheckBoxes[0] = new wxCheckBox(this, wxID_ANY, _T("Tile name"));
  m_pSizerChecks->Add(m_pArrCheckBoxes[0], 0, wxBOTTOM, vertSpaceChecks);
  m_pArrTextBoxes[0]  = new wxTextCtrl(this, wxID_ANY, _T(""), wxDefaultPosition, wxSize(160, 20));
  m_pSizerTextBoxes->Add(m_pArrTextBoxes[0], 0, wxBOTTOM, vertSpaceText);

  m_pArrCheckBoxes[1] = new wxCheckBox(this, wxID_ANY, _T("Tile label"));
  m_pSizerChecks->Add(m_pArrCheckBoxes[1], 0, wxBOTTOM, vertSpaceChecks);
  m_pArrTextBoxes[1]  = new wxTextCtrl(this, wxID_ANY, _T(""), wxDefaultPosition, wxSize(160, 20));
  m_pSizerTextBoxes->Add(m_pArrTextBoxes[1], 0, wxBOTTOM, vertSpaceText);

  m_pArrCheckBoxes[2] = new wxCheckBox(this, wxID_ANY, _T("North glue"));
  m_pSizerChecks->Add(m_pArrCheckBoxes[2], 0, wxBOTTOM, vertSpaceChecks);
  m_pArrTextBoxes[2]  = new wxTextCtrl(this, wxID_ANY, _T(""), wxDefaultPosition, wxSize(160, 20));
  m_pSizerTextBoxes->Add(m_pArrTextBoxes[2], 0, wxBOTTOM, vertSpaceText);

  m_pArrCheckBoxes[3] = new wxCheckBox(this, wxID_ANY, _T("East glue"));
  m_pSizerChecks->Add(m_pArrCheckBoxes[3], 0, wxBOTTOM, vertSpaceChecks);
  m_pArrTextBoxes[3]  = new wxTextCtrl(this, wxID_ANY, _T(""), wxDefaultPosition, wxSize(160, 20));
  m_pSizerTextBoxes->Add(m_pArrTextBoxes[3], 0, wxBOTTOM, vertSpaceText);

  m_pArrCheckBoxes[4] = new wxCheckBox(this, wxID_ANY, _T("South glue"));
  m_pSizerChecks->Add(m_pArrCheckBoxes[4], 0, wxBOTTOM, vertSpaceChecks);
  m_pArrTextBoxes[4]  = new wxTextCtrl(this, wxID_ANY, _T(""), wxDefaultPosition, wxSize(160, 20));
  m_pSizerTextBoxes->Add(m_pArrTextBoxes[4], 0, wxBOTTOM, vertSpaceText);

  m_pArrCheckBoxes[5] = new wxCheckBox(this, wxID_ANY, _T("West glue"));
  m_pSizerChecks->Add(m_pArrCheckBoxes[5], 0, wxBOTTOM, vertSpaceChecks);
  m_pArrTextBoxes[5]  = new wxTextCtrl(this, wxID_ANY, _T(""), wxDefaultPosition, wxSize(160, 20));
  m_pSizerTextBoxes->Add(m_pArrTextBoxes[5], 0, wxBOTTOM, vertSpaceText);

  if (dirNum == 6) {
    m_pArrCheckBoxes[6] = new wxCheckBox(this, wxID_ANY, _T("Up glue"));
    m_pSizerChecks->Add(m_pArrCheckBoxes[6], 0, wxBOTTOM, vertSpaceChecks);
    m_pArrTextBoxes[6]  = new wxTextCtrl(this, wxID_ANY, _T(""), wxDefaultPosition, wxSize(160, 20));
    m_pSizerTextBoxes->Add(m_pArrTextBoxes[6], 0, wxBOTTOM, vertSpaceText);

    m_pArrCheckBoxes[7] = new wxCheckBox(this, wxID_ANY, _T("Down glue"));
    m_pSizerChecks->Add(m_pArrCheckBoxes[7], 0, wxBOTTOM, vertSpaceChecks);
    m_pArrTextBoxes[7]  = new wxTextCtrl(this, wxID_ANY, _T(""), wxDefaultPosition, wxSize(160, 20));
    m_pSizerTextBoxes->Add(m_pArrTextBoxes[7], 0, wxBOTTOM, vertSpaceText);
  } else {
    m_pArrCheckBoxes[6] = NULL;
    m_pArrCheckBoxes[7] = NULL;
  }

  pSizerControls->Add(m_pSizerChecks, 0, wxLEFT, 20);
  pSizerControls->Add(m_pSizerTextBoxes, 0, wxLEFT, 10);
  m_pSizerMain->Add(pSizerControls);

  m_pSizerMain->AddSpacer(10);
  wxBoxSizer* pButtonSizer = new wxBoxSizer(wxHORIZONTAL);
  wxButton* pOK = new wxButton(this, wxID_OK, _T("OK"));
  pButtonSizer->Add(pOK, 0, wxLEFT, 140);
  wxButton* pCancel = new wxButton(this, wxID_CANCEL, _T("Cancel"));
  pButtonSizer->Add(pCancel, 0, wxRIGHT, 5);

  m_pSizerMain->Add(pButtonSizer);
  m_pSizerMain->AddSpacer(5);

  SetAutoLayout(TRUE);
  SetSizer(m_pSizerMain);

  m_pSizerMain->SetSizeHints(this);
  m_pSizerMain->Fit(this);
}

void TASFindTilesWindow::OnOk(wxCommandEvent& WXUNUSED(event))
{
  for (int i=0; i<8; i++) {
    if (m_pArrCheckBoxes[i] && m_pArrCheckBoxes[i]->IsChecked()) {
      m_arrSearchProperties[i]  = true;
      m_arrSearchStrings[i]     = m_pArrTextBoxes[i]->GetValue();
    } else {
      m_arrSearchProperties[i]  = false;
    }
  }

  EndModal(wxID_OK);
}

///////////////////////////////////////////////////////////////////////

void TASTileEditor::OnFind(wxCommandEvent& WXUNUSED(event))
{
  TASFindTilesWindow dlg(this);
  if (dlg.ShowModal() == wxID_OK) {
    m_selectedList.clear();

    int propCount = 0;
    for (int i=0; i<8; i++) {
      if (dlg.m_arrSearchProperties[i]) {
        propCount++;
      }
    }

    for (size_t i=0; i<g_pEditorTileset->size(); i++) {
      int propsMatched = 0;
      if (dlg.m_arrSearchProperties[0] && g_pEditorTileset->at(i)->m_sName.Contains(dlg.m_arrSearchStrings[0])) {
        propsMatched++;
      }
      if (dlg.m_arrSearchProperties[1] && g_pEditorTileset->at(i)->m_sLabel.Contains(dlg.m_arrSearchStrings[1])) {
        propsMatched++;
      }
      for (int j=0; j<6; j++) {
        if (dlg.m_arrSearchProperties[j+2] && g_pEditorTileset->at(i)->m_sides[j].m_sLabel.Contains(dlg.m_arrSearchStrings[j+2])) {
          propsMatched++;
        }
      }
      if (propsMatched == propCount) {
        m_selectedList.push_back(i);
      }
    }
    wxString sMsg = wxString::Format(_T("Found %d tiles matching the search string(s)\n"), m_selectedList.size());
    g_pTASApp->SetEditorOutputText(sMsg, true);
    Refresh();
  }
}

void TASTileEditor::ScrollToCenter(int tileNum)
{
  if (tileNum > -1) {
    int x, y;
    GetTileLocation(tileNum, &x, &y);

    int xScrollUnits, yScrollUnits;
    GetScrollPixelsPerUnit(&xScrollUnits, &yScrollUnits);

    int w, h;
    GetClientSize(&w, &h);
    y = y - (h / 2);

    int iXP = x / xScrollUnits;
    int iYP = y / yScrollUnits;

    Scroll(iXP, iYP);
    m_pEditorFrame->m_pTTWin->SetCurrentTileType(tileNum);
    SelectTile(tileNum);
    m_pEditorFrame->Raise();
  }
}

void TASTileEditor::FindTileType(wxString sName, bool bScrollToCenter)
{
  int tileNum = -1;
  m_selectedList.clear();
  for (size_t i=0; i<g_pEditorTileset->size(); i++) {
    if (g_pEditorTileset->at(i)->m_sName == sName) {
      m_selectedList.push_back(i);
      tileNum = i;
      break;
    }
  }
  if (bScrollToCenter) {
    ScrollToCenter(tileNum);
  }
  Refresh();
}

void TASTileEditor::OnFindBinding(wxCommandEvent& event)
{
  if (m_selectedList.size() <= 0) return;

  int dir = event.GetId() - ID_TAS_TILEED_FINDBINDING_N;
  TASIntVect selTiles = m_selectedList;
  m_selectedList.clear();

  for (size_t j=0; j<selTiles.size(); j++) {
    for (size_t i=0; i<g_pEditorTileset->size(); i++) {
      if (g_pEditorTileset->at(selTiles[j])->m_sides[dir] == g_pEditorTileset->at(i)->m_sides[(dir+2)%4]) {
        SelectTile(i);
      }
    }
  }
  wxString sMsg = wxString::Format(_T("Found %d tiles that can bind\n"), m_selectedList.size());
  g_pTASApp->SetEditorOutputText(sMsg, true);
  Refresh();
}

void TASTileEditor::OnSelectUsed(wxCommandEvent& event)
{
  if (g_pEditorTileset->GetSimMod()) {
    wxString sError;
    sError += _T("Error:  The tile set in the editor has been\n");
    sError += _T("modified and is different from the tile set\n");
    sError += _T("in the simulator.  This operation cannot be\n");
    sError += _T("performed unless they are identical.");
    wxMessageBox(sError, _T("Modified tile set"), wxICON_ERROR|wxOK);
    return;
  }

  bool bShowUsed = (event.GetId() == ID_TAS_TILEED_SELECTUSED ? true : false);

  m_selectedList.clear();
  TASTileTypeSet unusedTypes;
  g_pTileSpace->GetUnusedTileTypes(g_pTileSpace->m_epoch, &unusedTypes);

  if (unusedTypes.size() == 0) {
    wxString msg(_T("All tile types are currently being used.\n"));
    g_pTASApp->SetSimDebugText(msg, true);
  } else if (bShowUsed) {
    wxString msg = wxString::Format(_T("Currently there are %d tiles types being used:\n"), (g_pTileSpace->m_tileset.size() - unusedTypes.size()));
    g_pTASApp->SetSimDebugText(msg, true);
  } else {
    wxString msg = wxString::Format(_T("Currently there are %d unused tiles types:\n"), unusedTypes.size());
    g_pTASApp->SetSimDebugText(msg, true);
  }

  for (size_t i=0; i<g_pTileSpace->m_tileset.size(); i++) {

    bool bInList = (unusedTypes.IndexOf(g_pTileSpace->m_tileset.at(i)->m_sName) != TAS_TILE_NOT_FOUND);
    if (bShowUsed && !bInList) {
      m_selectedList.push_back(i);

      wxString msg = wxString::Format(_T("  Name: %s\t Label: %s\n"), g_pTileSpace->m_tileset.at(i)->m_sName.c_str(), g_pTileSpace->m_tileset.at(i)->m_sLabel.c_str());
      g_pTASApp->SetSimDebugText(msg, true);
    } else if (!bShowUsed && bInList) {
      m_selectedList.push_back(i);

      wxString msg = wxString::Format(_T("  Name: %s\t Label: %s\n"), g_pTileSpace->m_tileset.at(i)->m_sName.c_str(), g_pTileSpace->m_tileset.at(i)->m_sLabel.c_str());
      g_pTASApp->SetSimDebugText(msg, true);
    }
  }
  unusedTypes.clear();
  Refresh();
}

void TASTileEditor::OnKey(wxKeyEvent& event)
{
  if ((event.GetModifiers() == wxMOD_CONTROL) && (event.GetKeyCode() == WXK_TAB)) {
    m_pEditorFrame->m_pEditorFrameNB->SetSelection((m_pEditorFrame->m_pEditorFrameNB->GetSelection() + 1) % m_pEditorFrame->m_pEditorFrameNB->GetPageCount());
  } else if ((event.GetModifiers() == (wxMOD_CONTROL|wxMOD_SHIFT)) && (event.GetKeyCode() == WXK_TAB)) {
    if (m_pEditorFrame->m_pEditorFrameNB->GetSelection() == 0) {
      m_pEditorFrame->m_pEditorFrameNB->SetSelection(m_pEditorFrame->m_pEditorFrameNB->GetPageCount() - 1);
    } else {
      m_pEditorFrame->m_pEditorFrameNB->SetSelection((m_pEditorFrame->m_pEditorFrameNB->GetSelection() - 1) % m_pEditorFrame->m_pEditorFrameNB->GetPageCount());
    }
  } else {
    event.Skip();
  }
}

wxPoint ptZero(0, 0);

void TASTileEditor::DrawTileEx(wxDC* pDC, int tileNum, int xOffset, int yOffset)
{
  if (tileNum < 0 || tileNum >= (int)g_pEditorTileset->size()) return;

  int sX, sY;
  GetTileLocation(tileNum, &sX, &sY);

  sX = sX + xOffset;
  sY = sY + yOffset;

  int z = 0;
  wxPen pen(g_pContext->m_colSpaceBkgrnd);
  wxBrush brush(g_pContext->m_colSpaceBkgrnd);
  bool bIsSelected = false;
  for (size_t s=0; s<m_selectedList.size(); s++) {
    if (m_selectedList[s] == tileNum) {
      bIsSelected = true;
      break;
    }
  }
  if (bIsSelected) {
    brush.SetColour(*wxBLUE);
    z = 1;
  } else if (tileNum == m_currTileTypeNum) {
    brush.SetColour(*wxGREEN);
    z = 1;
  }
  pDC->SetPen(pen);
  pDC->SetBrush(brush);
  wxDouble x, y, w, h;
  x = z-(m_tileSpacing/2);
  y = z-(m_tileSpacing/2);
  if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
    w = (g_tileSize + m_tileSpacing - (2*z));
    h = (g_tileSize + m_tileSpacing - (2*z));
  } else {
    w = (m_size3DTiles.GetWidth() + m_tileSpacing - (2*z));
    h = (m_size3DTiles.GetHeight() + m_tileSpacing - (2*z));
  }
  pDC->DrawRectangle(x+sX, y+sY, w, h);

  int diff = 3;

  if (g_pEditorTileset->GetEquivalentTileType(g_pEditorTileset->at(tileNum)) != NULL) {
    brush.SetColour(*wxRED);
    pDC->SetBrush(brush);
    pDC->DrawRectangle(x+diff+sX, y+diff+sY, w-7, h-7);
  }

  wxPoint pt(sX+x+m_tileSpacing/2, sY+y+m_tileSpacing/2);
  if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
    g_pEditorTileset->at(tileNum)->DrawTile(pDC, pt, g_tileSize, g_tileSize, g_pContext->m_maskEditorTileDisplay);
  } else {
    g_pEditorTileset->at(tileNum)->DrawFlat3DTile(pDC, pt, m_size3DTiles);
  }
}

wxString TASTileEditor::GenerateTileName(const wxChar* cNameAttempt)
{
  wxString sNewName(cNameAttempt);
  int addNum = 2;
  bool bContinue = true;
  while (bContinue) {
    bool bFound = false;
    for (size_t i=0; i<g_pEditorTileset->size(); i++) {
      if (g_pEditorTileset->at(i)->m_sName == sNewName) {
        sNewName = wxString::Format(_T("%s(%d)"), cNameAttempt, addNum);
        bFound = true;
        addNum++;
        break;
      }
    }
    if (!bFound) {
      bContinue = false;
    }
  }
  return sNewName;
}

