/////////////////////////////////////////////////////////////////////////////
// Name:        TASTiles.h
// Purpose:     Declares classes for tiles, tile space, etc.
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef __TASTiles__
#define __TASTiles__

#include <vector>
#include <map>
#include <set>
#include <memory>
#include <iostream>

#include <wx/string.h>
#include <wx/ffile.h>
#include <wx/image.h>
#include <wx/dc.h>

#include "TAS.h"
#include "TASCuts.h"

// Needed for simulating ktam
#include "StochasticSimulator.h"
#include "CompositionAndRejection.h"
#include "KTamEvent.h"
#include "TileConcentrationEvent.h"

#define TAS_NORTH   0
#define TAS_EAST    1
#define TAS_SOUTH   2
#define TAS_WEST    3
#define TAS_UP      4
#define TAS_DOWN    5
extern int TASOppositeDirection[];
extern TASCoords coordsNbrhd[7];
wxString TASDirectionNumToString(int dirNum, bool bAllCaps = false);

#define TAS_EMPTY_TTYPE_STR         _T("<empty>")
#define TAS_FRONTIER_TTYPE_STR      _T("<frontier>")

#define TAS_2D_MODE             0
#define TAS_3D_MODE             1
#define TAS_ATAM_MODE           0
#define TAS_KTAM_MODE           1
#define TAS_2HAM_MODE           2
#define TAS_TILE_MODE           9
#define TAS_OPT_MODE_0          0
#define TAS_OPT_MODE_1          1
#define TAS_TILECONC_MODE_OFF   0
#define TAS_TILECONC_MODE_ON    1
#define TAS_TILE_ADD_ONE        0
#define TAS_TILE_ADD_FULL_FRONT 1
#define TAS_TILE_TIMING_MODEL   2

#define TAS_MODE_ATAM_STR           _T("aTAM")
#define TAS_MODE_KTAM_STR           _T("kTAM")
#define TAS_MODE_2HAM_STR           _T("2HAM")

#define TAS_2HAM_INFINITE_COUNT     (size_t)(-1)

#define TAS_SPACE_GROWTHINC_2D      200
#define TAS_SPACE_GROWTHINC_3D      40

#define TAS_TILE_NOT_FOUND          (size_t)(-1)
#define TAS_TILE_FRONTIER           (size_t)(-2)
bool TASIsTileType(size_t t);

#define TAS_2HAMHACK_INIT         0
#define TAS_2HAMHACK_DONT_DELETE  1
#define TAS_2HAMHACK_DELETE       2

extern int g_tileSize;

#define TAS_TILE_NO_BINDING     0
#define TAS_TILE_INPUT_BINDING  1
#define TAS_TILE_OUTPUT_BINDING 2

wxString TASGetModeName(int mode);

class TASTileBinding
{
public:
  TASTileBinding() { m_type = TAS_TILE_NO_BINDING; m_strength = 0; };

  int m_type;
  int m_strength;
};

class TASTileSide
{
public:
  TASTileSide() { m_iStrength = 0; m_sLabel = _T(""); m_hashNum = TAS_TILE_NOT_FOUND; };
  TASTileSide(int strength, wxString label) { m_iStrength = strength; m_sLabel = label; };
  bool operator==(const TASTileSide& right);
  bool operator!=(const TASTileSide& right) { return !operator==(right); };

  int       m_iStrength;
  wxString  m_sLabel;
  size_t    m_hashNum;
};

#define TAS_MASK_DISPLABEL      1
#define TAS_MASK_DISPNAME       2
#define TAS_MASK_DISPGLUES      4
#define TAS_MASK_DISPSTRENGTHS  8
#define TAS_MASK_DISPALL        (TAS_MASK_DISPLABEL | TAS_MASK_DISPNAME | TAS_MASK_DISPGLUES | TAS_MASK_DISPSTRENGTHS)

#define TAS_MASK_ISON(mask, dispobj) ((mask & dispobj) == dispobj)

class TASTileTypeBitmapHolder
{
public:
  TASTileTypeBitmapHolder() { m_height = (size_t)-1; m_displayMask = TAS_MASK_DISPALL; };

  wxBitmap  m_bitmap;
  size_t    m_height;
  unsigned  m_displayMask;
};

class TASTileType
{
public:
  TASTileType(int dims = 2);
  TASTileType(const TASTileType& tt);
  TASTileType(const wxString& sTTDefn,    //  This constructor takes a string representation of the tiletype
              size_t* index);             //  or multiple tiletypes and an index into the string denoting
                                          //  where the constructor should begin looking for the definintion,
                                          //  and updates that index value to point to the portion of the
                                          //  string following the current definition
  ~TASTileType();

  bool IsValid();
  bool IsEqual(TASTileType* pTtype, bool bCompareNameAndLabel = false, bool bAll = false);
  void SetEqual(const TASTileType& tt);
  
  bool Rotate(int degrees);         //  Rotation is done in the clockwise direction

  wxString Serialize();

  //  Functions for rendering tiles
  void DrawTile(wxDC* pDC, wxPoint& pt, size_t width, size_t height, size_t displayMask);
  void DrawBond(wxDC* pDC, wxPoint& pt, int dir, bool bShowGlues, bool bSVG = false);
  void DrawFlat3DTile(wxDC* pDC, wxPoint& pt, wxSize& sz);
  void DrawFlat3DSide(wxDC* pDC, wxPoint& pt, int dim, int side);
  wxBitmap& GetBitmap(size_t dim);
  void InvalidateBitmaps();
//MJP: not currently implemented due to issues with the svg library
//  void DrawSVG(wxDC* pDC, wxPoint& pt, int dim, unsigned displayMask); 

  int            m_dims;
  int            m_concentration;
  wxString       m_sName;
  wxString       m_sLabel;
  wxString       m_sComment;
  wxString       m_sImagePath;
  TASTileSide    m_sides[6];
  wxImage        m_image;
  wxColour       m_colTile;
  wxColour       m_colText;
  wxBitmap       m_bitmapFlat3D;

protected:
  wxPen          m_pen;
  wxBrush        m_brush;
  std::vector<TASTileTypeBitmapHolder>  m_vecBitmaps;
};

typedef std::multimap<size_t, TASTileType*, std::less<size_t>, std::allocator<std::pair<const size_t, 
                                               TASTileType*> > > multimaptype;
typedef multimaptype::value_type multimapvaluetype;

//  TASTileTypeVect creates copies of the TASTileTypes that are to be inserted
class TASTileTypeVect : public std::vector<TASTileType*>
{
public:
  ~TASTileTypeVect();
  void push_back(const TASTileType* pTT);
  void insert(TASTileTypeVect::iterator iter, const TASTileType* pTT);
  void erase(TASTileTypeVect::iterator iter);
  void clear();
};

//  TASTileTypePtrVect is very simply just a vector of pointers to TASTileTypes
class TASTileTypePtrVect : public std::vector<TASTileType*>
{
public:
};

class TASTileSideHash
{
public:
  TASTileSideHash() {};
  TASTileSideHash(TASTileSide& tsNorth, TASTileSide& tsEast,
                  TASTileSide& tsSouth, TASTileSide& tsWest,
                  TASTileSide& tsUp,    TASTileSide& tsDown);

  void CreateHash();

  bool operator<(const TASTileSideHash& tshRight) const;
  bool operator==(const TASTileSideHash& tshRight);

  TASTileSide   m_arr[6];
  wxString      m_strHash;
};

class TASTileSideHashMap : public std::map<TASTileSideHash, TASTileTypePtrVect>
{

};


#define TAS_DEFAULT_GLUE _T("")

class TASGlueFunction : public std::map<wxString, int>
{
public:
  TASGlueFunction(int defaultStrength = 0) { SetDefaultStrength(defaultStrength); };
  int GetStrengthWithGlue(wxString& strIn);
  int GetMaxStrength();
  int GetMinStrength();
  void SetDefaultStrength(int defaultStrength) { mDefaultStrength = defaultStrength; };

  int mDefaultStrength;
};

class TASGlueFunctionsMap : public std::map<wxString, TASGlueFunction>
{
public:
};

class TASTileTypeSet
{
public:
  TASTileTypeSet() { m_bUsesGlueFunctionMap = false; m_bUsesNegativeGlues = false; };
  TASTileTypeSet& operator=(const wxString& sTTDefns);
  TASTileTypeSet& operator=(TASTileTypeSet& tts);
  TASTileType* operator[](size_t ind);
  TASTileType* operator[](const wxString& sName);

  TASTileType* at(size_t ind);
  TASTileType* at(const wxString& sName);

  size_t LoadGlueFunctionDefinitions(wxString& sTTDefns, size_t index);
  bool SaveToFile(wxFFile* outFile);

  size_t size();
  bool push_back(const TASTileType& tt, bool bReIndex = true);
  bool insert(const TASTileType& tt, size_t ind);
  bool erase(size_t ind);
  void clear();

  size_t IndexOf(const wxString& sName);
  size_t IndexOf(const TASTileType& tt);

  TASTileTypePtrVect* GetTileTypeListFromHash(TASTileSideHash& tsh, int temp);
  TASTileType* GetEquivalentTileType(TASTileType* pTtype);
  int GetUnusedTileTypes(int epoch, TASTileTypeSet* pUnusedTypes);
  int GetGlueInteractionStrength(TASTileSide& ts1, TASTileSide& ts2);
  int GetPossibleGlueInteractionStrength(TASTileSide& ts, bool bMax = true);

  //  This is used for detecting functionally equivalent tile types
  std::map<size_t, TASTileTypePtrVect> m_mmapGlues[6];

  TASTileSideHashMap m_mapTileSides;

  void InvalidateBitmaps();

  virtual void ReIndex();

  bool                        m_bUsesNegativeGlues;

protected:
  virtual void IndexTile(TASTileType* pTT, size_t num);

  TASTileTypeVect             m_vectTypes;
  std::map<wxString, size_t>  m_mapNameIndex;
  std::map<wxString, size_t>  m_mapGlueIndex;
  TASGlueFunctionsMap         m_mapGlueFunctions;
  bool                        m_bUsesGlueFunctionMap;
};

inline int TASGlueFunction::GetStrengthWithGlue(wxString& strIn)
{
  TASGlueFunction::iterator it = this->find(strIn);
  if (it == end()) {
    return mDefaultStrength;
  }
  return it->second;
}

inline int TASGlueFunction::GetMaxStrength()
{
  int ret = 0;
  for (TASGlueFunction::iterator it = begin(); it != end(); it++) {
    if (it->second > ret) {
      ret = it->second;
    }
  }
  return ret;
}

inline int TASGlueFunction::GetMinStrength()
{
  int ret = 0;
  for (TASGlueFunction::iterator it = begin(); it != end(); it++) {
    if (it->second < ret) {
      ret = it->second;
    }
  }
  return ret;
}

inline int TASTileTypeSet::GetGlueInteractionStrength(TASTileSide& ts1, TASTileSide& ts2) {
  if (m_bUsesGlueFunctionMap) {
    TASGlueFunctionsMap::iterator it = m_mapGlueFunctions.find(ts1.m_sLabel);
    if (it != m_mapGlueFunctions.end()) {
      return it->second.GetStrengthWithGlue(ts2.m_sLabel);
    }
    it = m_mapGlueFunctions.find(ts2.m_sLabel);   //  Check for the glue function definition in either direction,
                                                  //  e.g. "glueX + glueY = 2" OR "glueY + glueX = 2"
    if (it != m_mapGlueFunctions.end()) {
      return it->second.GetStrengthWithGlue(ts1.m_sLabel);
    }
    if (ts1 == ts2) {                             //  Assume diagonal strength function by default
      return ts2.m_iStrength;
    }
  } else if (ts1 == ts2) {
    return ts2.m_iStrength;
  }
  return 0;
}

inline int TASTileTypeSet::GetPossibleGlueInteractionStrength(TASTileSide& ts, bool bMax)
{
  if (m_bUsesGlueFunctionMap) {
    TASGlueFunctionsMap::iterator it = m_mapGlueFunctions.find(ts.m_sLabel);
    if (it != m_mapGlueFunctions.end()) {
      if (bMax) {
        return it->second.GetMaxStrength();
      } else {
        return it->second.GetMinStrength();
      }
    } else {
      int ret = (bMax ? ts.m_iStrength : 0);    //  Assume diagonal strength function by default
      //  MJP: the following loop will be an efficient way of searching the glue functions map as long as it's small.
      //       Otherwise, we may want to consider building up a reciprocal version of it as it is loaded
      for (TASGlueFunctionsMap::iterator it = m_mapGlueFunctions.begin(); it != m_mapGlueFunctions.end(); it++) {
        for (TASGlueFunction::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++) {
          if (it2->first == ts.m_sLabel) {
            if (bMax && (ret < it2->second)) {
              ret = it2->second;
            } else if (!bMax && (ret > it2->second)) {
              ret = it2->second;
            }
          }
        }
      }
      return ret;
    }
  } else {
    return (bMax ? ts.m_iStrength : 0);
  }
}


class TASTileTypeSetEx : public TASTileTypeSet
{
public:
  TASTileTypeSetEx& operator=(const wxString& sTTDefns);

  void IndexTile(TASTileType* pTT, size_t num);
  void ReIndex();

  void SetDiskMod(bool bIsMod)  { m_bModifiedFromDisk = bIsMod; };
  void SetSimMod(bool bIsMod)   { m_bModifiedFromSim  = bIsMod; };
  bool GetDiskMod()             { return m_bModifiedFromDisk; };
  bool GetSimMod()              { return m_bModifiedFromSim; };

  wxArrayString                 m_sArrNames;
  wxArrayString                 m_sArrLabels;
  wxArrayString                 m_sArrGlues;
  std::map<wxString, int>       m_sMapLabels;
  std::map<wxString, int>       m_sMapGlues;

protected:
  bool                          m_bModifiedFromDisk;
  bool                          m_bModifiedFromSim;
};

class TASFrontierLocation
{
public:
  TASFrontierLocation() {};
  TASFrontierLocation(const TASCoords& coords) { m_coords = coords; };
  
  inline bool operator<(const TASFrontierLocation& tfl) { return (m_coords < tfl.m_coords); };
  inline bool operator<(const TASFrontierLocation& tfl) const { return (m_coords < tfl.m_coords); };

  TASCoords   m_coords;
};

class TASFrontierSet : public std::set<TASFrontierLocation>
{
public:
  TASFrontierLocation* GetElement(size_t ind);
  bool RemoveElement(const TASCoords& coords);
};

class TileLocationIndexItem
{
public:
  TileLocationIndexItem() { m_tiletype = TAS_TILE_NOT_FOUND; m_epoch = 0; };
  TileLocationIndexItem(size_t tiletype, size_t epoch) { m_tiletype = tiletype; m_epoch = epoch; };

  size_t  m_tiletype;
  size_t  m_epoch;
};

//  The TileLocationIndex contains a value for every possible tile location in the current space.
//  The value for each location is one of the following:
//    - for locations with a tile, the tileType number
//    - for locations without a tile but on the frontier, TAS_TILE_FRONTIER
//    - for all other locations, TAS_TILE_NOT_FOUND
class TileLocationIndex
{
public:
  TileLocationIndex();
  ~TileLocationIndex();

  size_t GetSize() { return m_pLocs->size(); };
  size_t GetLocation(size_t ind);
  size_t GetEpoch(size_t ind);
  void SetLocation(size_t ind, size_t tileTypeNum, size_t epoch);
  void Clear();

protected:
  std::vector<TileLocationIndexItem>*  m_pLocs;
  size_t                m_growthFactor;
};

//  An instance of the class TASLocationTransition keeps track of one
//  transition from one of the following types to another at a given locaiton:
//  0 - maxTileTypeNum (the number representing some tile type)
//  TAS_TILE_NOT_FOUND
//  TAS_TILE_FRONTIER
class TASLocationTransition
{
public:
  TASLocationTransition() { m_typeBegin = TAS_TILE_NOT_FOUND; m_typeEnd = TAS_TILE_NOT_FOUND; };
  TASLocationTransition(const TASLocationTransition& trans) { m_coords = trans.m_coords; m_typeBegin = trans.m_typeBegin; m_typeEnd = trans.m_typeEnd; };
  TASLocationTransition(const TASCoords& coords, size_t typeStart, size_t typeEnd) { m_coords = coords; m_typeBegin = typeStart; m_typeEnd = typeEnd; };

  bool operator==(const TASLocationTransition& trans) { return (m_coords == trans.m_coords && m_typeBegin == trans.m_typeBegin && m_typeEnd == trans.m_typeEnd); };
  TASLocationTransition& operator=(const TASLocationTransition& trans) { m_coords = trans.m_coords; m_typeBegin = trans.m_typeBegin; m_typeEnd = trans.m_typeEnd; return *this; };

  TASCoords       m_coords;
  size_t          m_typeBegin;
  size_t          m_typeEnd;
};
WX_DECLARE_OBJARRAY(TASLocationTransition, TASLocationTransitionList);

//  An instance of TASTileEpoch stores the set of transitions that occurred
//  in an epoch.  The transitions are stored in a single list which is
//  grouped by the types of the ends of the transitions, in the following order:
//  <tile types>
//  TAS_TILE_FRONTIER
//  TAS_TILE_NOT_FOUND
enum base_types { TAS_BEGIN_TILES, TAS_BEGIN_FRONTIER, TAS_BEGIN_EMPTY };
class TASTileEpoch
{
public:
  TASTileEpoch();
  TASTileEpoch(int temp);

  void ConstructorHelper();
  bool AddTransition(TASLocationTransition& trans);
  bool RemoveTransition(TASLocationTransition& trans);
  //bool RemoveTransitionsToType(int baseType);
  bool RemoveTransitionAtLocation(TASCoords& coords);
  void ClearTransitions();
  size_t GetEnd();

  TASLocationTransitionList   m_listTransitions;
  int                         m_arrBases[4];
  int                         m_temp;
};
class TASTileEpochList : public std::vector<TASTileEpoch> {};

class TASSeedEntry
{
public:
  TASSeedEntry(wxString sName, TASCoords& coords) { m_sName = sName; m_coords = coords; };

  wxString    m_sName;
  TASCoords   m_coords;
};

class TASSeed : public std::vector<TASSeedEntry>
{
public:
  TASSeed() { m_iAvgX = 0; m_iAvgY = 0; };

  TASSeed& operator=(TASSeed& tts);

  int       m_iAvgX;
  int       m_iAvgY;
  int       m_iAvgZ;
  TASCoords m_coordsLeftmost;
};

class TASTileSpace : public TASSpace
{
  friend class TASCutSets;
  friend class TASTilePanel;
public:
  TASTileSpace(int dimNum = 2);
  virtual ~TASTileSpace();

  void ConstructorHelper(int dimNum);

  virtual bool IncrementEpoch();
  virtual bool DecrementEpoch();
  virtual bool IsEpochCached(size_t epoch) { return m_listEpochCache.size() >= (epoch+1); };
  void AddEpoch(bool bDoInc = false);

  bool AddTile(const TASCoords& coords, const TASTileType* pTT);
  bool RemoveTile(const TASCoords& coords);

  bool SetTileTypes(wxString& sTileTypeDefns);
  bool SetTileTypes(TASTileTypeSet& setIn);

  inline int GetTemp() { return m_temp; };
  bool SetTemp(int temp);
  size_t AttemptMelting(TileLocationIndex* pPossibleLocsIn);

  void RebuildFrontier();

  void SetTerminal(bool bIn) { m_bIsTerminal = bIn; };
  bool IsTerminal() { return m_bIsTerminal; };
  bool IncrementStage();

  size_t Index(const TASCoords& coords);
  TASCoords CoordsFromIndex(size_t ind);

  inline int GetXDim() { return m_coordsDims.x; };
  inline int GetYDim() { return m_coordsDims.y; };
  inline int GetZDim() { return m_coordsDims.z; };

  inline int GetDimNum() { return m_dimNum; };
  void SetDimNum(int dimNum);
  inline int GetNumNbrs() { return m_numNbrs; };

  size_t GetRandSeed();
  void   SetRandSeed(unsigned long seed, bool bRandomSeed = false);

  size_t GetMaxEpoch();

  size_t GetNumTransitionsInEpoch(size_t epoch);
  TASLocationTransition* GetTransitionInEpoch(size_t epoch, size_t changeNum);

  //  The endType parameter for the following two functions should be one of the following:
  //    TAS_BEGIN_TILES, TAS_BEGIN_FRONTIER, TAS_BEGIN_EMPTY
  size_t GetNumTransitionsToTypeInEpoch(size_t epoch, int endType);
  TASLocationTransition* GetTransitionToTypeInEpoch(size_t epoch, size_t changeNum, int endType);

  virtual TASTileType* SafelyGetTileType(const TASCoords& coords);
  virtual bool IsFrontierLocation(const TASCoords& coords);
  bool KillFrontierLocation(TASFrontierLocation* pFrontLoc);
  virtual size_t GetEpochAdded(const TASCoords& coords);

  int GetUnusedTileTypes(int epoch, TASTileTypeSet* pUnusedTypes);

//MJP: could optimize saving and loading by also saving frontier so tiles don't have to be added one at a time with full calculations
  bool LoadSpaceDefinitionFromFile(wxFFile* inFile);
  virtual bool LoadSpaceDefinitionFromString(wxString sSeedDefn);
  bool SaveToFile(wxFFile* outFile);

  bool LoadStageDefinitionsFromFile(wxFFile* inFile);

  bool IsInSeed(const TASCoords& c);
  void ClearSeed();
  wxString ValidateSeed();                  //  If the seed is valid, the return string is empty.  Otherwise, it
                                            //  is the first tile type name not found in the current tile set
  bool LoadSeed();
  virtual void ResetSpace(bool bSaveSeed = false);

  void OutputChangesPerEpoch();
  void OutputChangesInEpoch(int epoch);

  void TruncateEpochCache(size_t maxEpoch);

//MJP: this group of functions can probably be much more optimized
  TASTileTypePtrVect* GetFittingTileTypes(const TASCoords& coords);
  TASTileTypePtrVect* GetFittingTileTypesRestricted(const TASCoords& coords, std::vector<TASTileBinding>* pVect);
  int GetIncidentGlueStrengthSum(const TASCoords& coords);
  bool AddTileToLocation(TASCoords coords, int typeInd = -1);
  bool TileCanBind(TASTileType* pTtype, const TASCoords& coords);
  int GetTileBindings(const TASCoords& coords, std::vector<TASTileBinding>* pVect);  // returns lowest binding value (to help check for negative cuts)
  void CheckForLocalDeterminism(TASCoords coords, TASTileTypePtrVect* pTTVect);
  bool AddTileToFrontier();
  bool AddMaxTilesToFrontier();
  bool UseTimingModelToAddTilesToFrontier();
  bool AddNeighborsToFrontierList(const TASCoords& coords);
  bool FixFrontierInNeighborhood(const TASCoords& coords);
  void CalculateTotalTileConcentrations();

  TASTileTypeSet        m_tileset;
  TASCoords             m_coordsMin;
  TASCoords             m_coordsMax;
  TASCoords             m_coordsDims;           //  Pre-calculated dimension sizes for efficiency
  TASFrontierSet        m_setFrontier;          //  Set of current frontier locations
  size_t                m_epoch;
  size_t                m_numTiles;
  bool                  m_bEpochCached;
  TASSeed               m_seed;
  std::vector<int>      m_vectTemps;            //  If temperature programming is being used, this will
                                                //  be the list of temperatures to apply
  // (Michael) This gets changed to negative one at some point
  //           so to preserve behavior I changed it to an int
  int                   m_currTemp;             //  The index into m_vectTemps of the temperature currently
                                                //  being used
  bool                  m_bUseStages;           //  Whether or not staged assembly is being used
  std::vector<std::set<wxString> > m_vectStages; //  If staged assembly is being used, this will be list of valid tile
                                                //  types for each stage
  size_t                m_currStage;            //  The index into m_vectStages of the stage currently being used
  TileLocationIndex*    m_pLocationIndex;       //  Effectively an n-dimensional array to directly index all current space locations
  TASCutSets            m_cutSets;              //  Set of cuts across the assembly that are below tau and can be discarded
  int                   m_iTotalConc;           //  The summed total of the concentration values of all tile types

protected:
  virtual TASTileType* GetTileType(const TASCoords& coords);
  virtual void GrowSpace(const TASCoords& coords);
  void SetDims(const TASCoords& coordsMin, const TASCoords& coordsMax);   //  Deletes existing space and sets the requested dimensions

  TASCoords CoordsFromIndexHelper(size_t ind, const TASCoords& dims, const TASCoords& mins);
  void SetSeed();                                 //  Sets the seed equal to all the tiles in epoch 0

  bool AddTileToFrontierHelper(int numTilesMode); //  Adds tiles to the frontier.  Options are TAS_TILE_ADD_ONE (add a single tile),
                                                  //  TAS_TILE_ADD_FULL_FRONT (fill the frontier), or TAS_TILE_ADD_APPROX_HALF (fill
                                                  //  about half of the frontier locations based on a normal distrubution over the frontier
                                                  //  size

  bool                    m_bIsTerminal;
  int                     m_dimNum;               //  Is space 2-D or 3-D (MJP: currently hardcoded to 2-D)
  int                     m_numNbrs;              //  The number of neighbors that a location has (depends on dimension setting)
  int                     m_temp;                 //  The temperature specifier for the number of valid bonds necessary for a tile to stick
  int                     m_dimGrowthRate;
  TASTileEpochList        m_listEpochCache;

  TASCoords               m_nbrOffsets[6];
};

typedef std::vector<TASTileType*> TASSupertileColumn;

class TASSupertile {
public:
  TASSupertile(TASTileType* pTT, TASTileTypeSet* pTS, size_t count = TAS_2HAM_INFINITE_COUNT);
  TASSupertile(TASSupertile* pST, TASTileTypeSet* pTS, size_t count = TAS_2HAM_INFINITE_COUNT);
  TASSupertile(TASSupertile* pST1, TASSupertile* pST2, TASCoords& offsetST2, TASTileTypeSet* pTS, size_t count = TAS_2HAM_INFINITE_COUNT);
  TASSupertile(TASSeed* pSeed, TASTileTypeSet* pTS, size_t count = TAS_2HAM_INFINITE_COUNT);

  void CalculateDims();
  bool ComboAttempted(TASSupertile* pST);
  bool GetCombinations(TASSupertile* pST, std::vector<TASSupertile*>* pVectRet);
  TASSupertile* CanCombine(size_t xTrans, size_t yTrans, TASSupertile* pST);
  bool IsEqual(TASSupertile* pST);

  void GenerateUID();
  void PrintToString(wxString* pStr);

  std::vector<TASSupertileColumn> m_vectSTC;
  wxString  m_sUID;

  TASTileTypeSet*         m_pTS;
  size_t                  m_minToCheck;   // The number of the first supertile that this supertile should be checked for combinations
  std::set<TASSupertile*> m_comboSet;     //  The set of supertiles which we've already attempted to combine with this one
  std::set<TASSupertile*> m_setChildren;  //  The set of supertiles (immediately) created from this one

  size_t            m_width;
  size_t            m_height;
  size_t            m_maxDim;
  size_t            m_numTiles;
  size_t            m_count;
  size_t            m_supertileNum;
  size_t            m_epochAdded;
};

class TASSupertileSpace {
public:
  TASSupertileSpace(size_t dims) { m_pST = NULL; m_dims = dims; };
  TASSupertileSpace(TASSupertile* pST, size_t dims) { m_pST = pST; m_dims = dims; CalculateOffsets(); };
  ~TASSupertileSpace();

  void CalculateOffsets();
  TASTileType* GetTileType(const TASCoords& coords);
  TASSupertile* GetSupertile(const TASCoords& coords);

  TASSupertile* m_pST;
  size_t        m_dims;
  size_t        m_offsetX;
  size_t        m_offsetY;
};

//  Forward declaration
class TAS2HAMThread;

class TAS2HAMEpochCacheEntry
{
public:
  TAS2HAMEpochCacheEntry(size_t maxST = 0, size_t numTiles = 0) { m_maxSupertile = maxST; m_numTilesInLargestST = numTiles; };

  size_t  m_maxSupertile;
  size_t  m_numTilesInLargestST;
};

class TAS2HAMTileSpace : public TASTileSpace
{
public:
  TAS2HAMTileSpace(int dimNum = 2);
  TAS2HAMTileSpace(TASTileSpace* pTileSpace);
  ~TAS2HAMTileSpace();

  void ResetSpace(bool bSaveSeed = false);
  void GrowSpace(const TASCoords& WXUNUSED(coords)) {};  // Don't need this for 2HAM space
  void AddSupertileSpace(TASSupertileSpace* pSTS, size_t epochAdded);
  void RemoveSupertileSpace(TASSupertileSpace* pSTS);
  TASSupertile* IsDuplicate(TASSupertile* pTS);

  bool IsFrontierLocation(const TASCoords& WXUNUSED(coords)) { return false; };
  size_t GetEpochAdded(const TASCoords& WXUNUSED(coords)) { return 0; };
  bool IsEpochCached(size_t epoch) { return m_list2HAMEpochCache.size() >= (epoch+1); };
  
  bool CreateWorkerThreads();
  bool DeleteWorkerThreads();
  bool IncrementEpoch();
  bool DecrementEpoch();

  //  Versions of tile and supertile accessors that are safe
  //  to use from outside of this class (e.g. froom UI threads)
  TASTileType* SafelyGetTileType(const TASCoords& coords);
  TASSupertile* SafelyGetSupertile(const TASCoords& coords);
  TASSupertileSpace* SafelyGetSupertileSpace(const TASCoords& coords);
  size_t SafelyGetNumTilesInLargestSupertile();

  void FindTerminalSupertiles();
  void CalculateStats(bool bForce = false, size_t epochIn = (size_t)-1);

  std::vector<TASSupertileSpace*>   m_vectSTS;
  std::vector<TASSupertileSpace*>   m_vectVisSTS;
  std::vector<TAS2HAMThread*>       m_vectWorkerThreads;
  std::map<wxString,TASSupertile*>  m_setST;
  wxSemaphore                       m_semCounterBlock;
  wxSemaphore                       m_semMainBlock;
  wxSemaphore                       m_semFreeCount;
  size_t                            m_currThreadMin;
  size_t                            m_currThreadMax;
  size_t                            m_numSupertilesBefore;
  size_t                            m_maxNewSupertile;
  std::vector<size_t>               m_vectStats;

  bool    m_bIncludeSingletons;
  size_t  m_numRows;
  size_t  m_numCols;
  size_t  m_maxSupertile;
  size_t  m_numSupertiles;
  size_t  m_maxVisSupertile;
  size_t  m_numVisSupertiles;
  size_t  m_defaultPadding;
  size_t  m_epochLastStats;

  size_t  m_minVisSize;
  size_t  m_maxVisSize;
  bool    m_bOnlyLargestVis;

  size_t  m_minKeepSize;
  size_t  m_maxKeepSize;
  bool    m_bOnlyLargestKeep;
protected:

  std::vector<TAS2HAMEpochCacheEntry> m_list2HAMEpochCache;
  size_t  m_numTilesInLargestST;

  TASTileType* GetTileType(const TASCoords& coords);
  TASSupertile* GetSupertile(const TASCoords& coords);
  TASSupertileSpace* GetSupertileSpace(const TASCoords& coords);
};

class TASkTAMTileSpace : public TASTileSpace {
public:
	TASkTAMTileSpace(int dimensions = 2) : TASTileSpace(dimensions) {
		gse = 0;
		gmc = 0;
		totalTime = 0.0;
        sim = NULL;
        errorCount = 0;
	}

	virtual ~TASkTAMTileSpace() {
        if (sim) {
              delete sim;
        }
	}

	virtual bool IncrementEpoch();
    
	// We can't reverse yet
	virtual bool DecrementEpoch() {
		return false;
	}
    
	virtual bool IsEpochCached(size_t WXUNUSED(epoch)) {
		// return m_listEpochCache.size() >= (epoch+1);
		return false;
	}

	// We don't keep a frontier
	virtual bool IsFrontierLocation(const TASCoords& WXUNUSED(coords)) {
		return false;
	}

	int GetBondStrength(const TASCoords& coords);
    bool ErrorAt(const TASCoords& coords);
	void UpdateTileEvents(const TASCoords& coords);
	bool HasNeighbor(const TASCoords& coords);
	
	bool HasMismatchedGlue(const TASCoords& coords);
	int GetErrorCount(void) {
		return errorCount;
	}

public:
	// Needed for configuration
	void setGmc(double gmc) { this->gmc = gmc; }
	void setGse(double gse) { this->gse = gse; }

	void initializeTileChooser();
	void initializeStochasticSimulator();

private:
	double totalTime;
	double gmc;
	double gse;
    
    int errorCount;

	double totalConcentration;

	StochasticSimulator<TileConcentrationEvent>* tileChooser;
	StochasticSimulator<KTamEvent>* sim;

	std::set<TASCoords> errors;
};

#endif /* __TASTiles__ */

